from hgc import constants

import wadi as wd

def test_correct_units():
    """ Test whether the to-hgc method is using the correct units and names
    of quantitties """
    assert True
    # assert constants == wd.used_hgc_units()




